from .accelerate import *
from .base import *
from .ddp import *
from .deepspeed import *
from .dp import *
from .utils import *
from typing import Literal


# 这里理论上不应该从Trainer继承，这里只是为了后续的代码提示
class AutoTrainer(Trainer):
    def __new__(cls, *args, trainer_type:Literal['deepspeed', 'ddp', 'dp', 'accelerate', 'auto', 'base']='base', **kwargs) -> Trainer:
        from torch4keras.snippets import is_deepspeed_available, is_accelerate_available, log_info
        if trainer_type == 'auto':
            if is_deepspeed_available():
                trainer_type = 'deepspeed'
            elif is_accelerate_available():
                trainer_type = 'accelerate'
            elif torch.cuda.device_count() > 1:
                if int(os.environ.get("RANK", -1)) != -1:
                    trainer_type = 'ddp'
                else:
                    trainer_type = 'dp'
            else:
                trainer_type = 'base'

        if trainer_type == 'deepspeed':
            trainer = DeepSpeedTrainer(*args, **kwargs)
        elif trainer_type == 'ddp':
            trainer = DDPTrainer(*args, **kwargs)
        elif trainer_type == 'dp':
            trainer = DPTrainer(*args, **kwargs)
        elif trainer_type == 'accelerate':
            trainer = AccelerateTrainer(*args, **kwargs)
        elif trainer_type == 'base':
            trainer = Trainer(*args, **kwargs)
        else:
            raise ValueError(f'Args `{trainer_type}` not supported')
        
        # log_info(f'Initialize `{trainer_type}` trainer success.')
        return trainer


from torch4keras.snippets import timeit, Timeit, Timeit2
import time


# 装饰器
@timeit
def func(n=10):
    for _ in range(n):
        time.sleep(0.1)
func()

# 上下文管理器 - 统计累计耗时
with Timeit() as ti:
    for i in range(10):
        time.sleep(0.1)
        ti.lap(name=i, reset=False)  # 统计累计耗时

# 上下文管理器 - 统计每段速度
with Timeit() as ti:
    for i in range(10):
        time.sleep(0.1)
        ti.lap(count=10, name=i, reset=True)
    ti(10) # 统计速度


# 上下文管理器 - 统计速度
with Timeit() as ti:
    for i in range(10):
        time.sleep(0.1)
        ti.lap(name=i, reset=True)
    ti(10) # 统计速度

ti = Timeit2()
for i in range(10):
    time.sleep(0.1)
    ti.lap(name=i)

for i in range(10):
    time.sleep(0.1)
    ti.lap(name=i)
ti.end() # 打印时长




from typing import Any, Callable, Dict, List, Optional, Tuple, Union
import time
import os
import traceback
import copy
import functools
from .log import log_info, log_warn, log_error
from pprint import pprint
import datetime


def format_timestamp(timestamp, format='%Y-%m-%d %H:%M:%S', verbose=0):
    '''格式化显示时间戳'''
    dt_object = datetime.datetime.fromtimestamp(timestamp)  
    
    # 格式化 datetime 对象  
    formatted_time = dt_object.strftime(format)  
    
    if verbose > 0:
        print(formatted_time)
    return formatted_time


def format_time(eta:Union[int, float], hhmmss=True):
    '''格式化显示时间间隔
    :param hhmmss: bool, 是否只以00:00:00格式显示
    '''
    # 以00:00:00格式显示
    if hhmmss:
        if eta > 86400:  # 1d 12:10:36
            eta_d, eta_h = eta // 86400, eta % 86400
            eta_format = '%dd ' % eta_d + ('%d:%02d:%02d' % (eta_h // 3600, (eta_h % 3600) // 60, eta_h % 60))
        elif eta > 3600:  # 12:10:36
            eta_format = ('%d:%02d:%02d' % (eta // 3600, (eta % 3600) // 60, eta % 60))
        else:  # 10:36
            eta_format = '%d:%02d' % (eta // 60, eta % 60)

    else:
        if eta > 86400:  # 1d 12h 10m 36s
            eta_d, eta_h = eta // 86400, eta % 86400
            eta_format = '%dd %dh %dm %ds' % (eta_d, eta_h // 3600, (eta_h % 3600) // 60, eta_h % 60)
        elif eta > 3600:  # 12h 10m 36s
            eta_format = ('%dh %dm %ds' % (eta // 3600, (eta % 3600) // 60, eta % 60))
        elif eta > 60:  # 10m 36s
            eta_format = ('%dm %ds' % (eta // 60, eta % 60))
        elif (eta >= 1) and (eta <= 60):  # 36.02s
            eta_format = '%.2fs' % eta
        elif eta >= 1e-3:  # 25ms
            eta_format = '%.0fms' % (eta * 1e3)
        else:  # 250us
            eta_format = '%.0fus' % (eta * 1e6)
    return eta_format


def timeit(func):
    '''装饰器, 计算函数消耗的时间
    
    Examples:
    ```python
    >>> @timeit
    >>> def main(n=10):
    ...     for i in range(n):
    ...         time.sleep(0.01)

    >>> main(10)
    ```
    '''
    def warpper(*args, **kwargs):
        start = time.time()
        res = func(*args, **kwargs)
        end = time.time()
        consume = format_time(end - start, hhmmss=False)
        start1 = time.strftime("%Y-%m-%d %H:%M:%S", time.localtime(start))
        end1 = time.strftime("%Y-%m-%d %H:%M:%S", time.localtime(end))

        log_info(f'Function `{func.__name__}` cost {consume} [{start1} < {end1}]')
        return res
    return warpper


class Timeit:
    '''上下文管理器, 记录耗时/平均耗时

    Examples:
    ```python
    >>> from torch4keras.snippets import Timeit
    >>> with Timeit() as ti:
    ...     for i in range(10):
    ...         time.sleep(0.1)
    ...         # ti.lap(name=i, reset=False)  # 统计累计耗时
    ...         # ti.lap(name=i, reset=True)  # 统计间隔耗时
    ...         # ti.lap(count=10, name=i, restart=True)  # 统计每段速度
    ...     # ti(10) # 统计速度
    ```
    '''
    def __enter__(self, template='Average speed: {:.2f}/s'):
        self.count = None
        self.start_tm = time.time()
        self.template = template
        return self

    def __call__(self, count):
        self.count = count

    def reset(self):
        '''自定义开始记录的地方'''
        self.start_tm = time.time()
    
    def lap(self, name:str=None, count:int=None, reset=False):
        '''
        :params name: 打印时候自定义的前缀
        :params count: 需要计算平均生成速度中统计的次数
        :params reset: 是否重置start_tm, True只记录时间间隔, 否则记录的是从一开始的累计时间
        '''
        if count is not None:
            self.count = count
        name = '' if name is None else str(name).strip() + ' - '

        end_tm = time.time()
        consume = end_tm - self.start_tm
        if self.count is None:
            # 只log时间
            consume = format_time(consume, hhmmss=False)
            start1 = time.strftime("%Y-%m-%d %H:%M:%S", time.localtime(self.start_tm))
            end1 = time.strftime("%Y-%m-%d %H:%M:%S", time.localtime(end_tm))
            log_info(name + f'Cost {consume} [{start1} < {end1}]')
        elif consume > 0:
            speed = self.count / consume
            log_info(name + self.template.format(speed))
        else:
            pass
            # log_warn('Time duration = 0')
        
        if reset:
            self.reset()

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.lap()
        print()


class Timeit2:
    '''记录耗时

    Examples:
    ```python
    >>> ti = Timeit2()
    >>> for i in range(10):
    ...     time.sleep(0.1)
    ...     ti.lap(name=i)
    >>> ti.end() # 打印各个步骤时长
    ```
    '''
    def __init__(self):
        self.reset()

    def __call__(self, *args, **kwargs):
        self.lap(*args, **kwargs)

    def reset(self):
        '''自定义开始记录的地方'''
        self.cost = dict()
        self.count = dict()
        self.start_tm = time.time()

    def restart(self):
        self.start_tm = time.time()

    def lap(self, name:str, verbose=0):
        '''
        :params name: 打印时候自定义的前缀
        '''
        end_tm = time.time()
        consume = end_tm - self.start_tm
        name = str(name)
        self.cost[name] = self.cost.get(name, 0) + consume
        self.count[name] = self.count.get(name, 0) + 1
        self.start_tm = time.time()
        
        if verbose > 1:
            start1 = time.strftime("%Y-%m-%d %H:%M:%S", time.localtime(self.start_tm))
            end1 = time.strftime("%Y-%m-%d %H:%M:%S", time.localtime(end_tm))
            log_info(name + f'Cost {consume} [{start1} < {end1}]')

    def end(self, verbose=1):
        for k, v in self.count.items():
            if v > 1:
                self.cost['avg_' + k] = self.cost[k] / v
        
        if verbose > 0:
            log_info(f'Cost detail: {self.cost}')
            self.reset()
        else:
            cost = copy.deepcopy(self.cost)
            self.reset()
            return cost


def send_email(mail_receivers:Union[str,list], mail_subject:str, mail_msg:str="", mail_host:str=None, 
               mail_user:str=None, mail_pwd:str=None, mail_sender:str=None):
    ''' 发送邮件(默认使用笔者自己注册的邮箱, 若含敏感信息请使用自己注册的邮箱)

    :param mail_subject: str, 邮件主题
    :param mail_msg: str, 邮件正文
    :param mail_receivers: str/list, 收件人邮箱
    :param mail_host: str, 发件服务器host
    :param mail_user: str, 发件人
    :param mail_pwd: str, smtp的第三方密码
    :param mail_sender: str, 发件人邮箱
    '''
    import smtplib
    from email.mime.text import MIMEText

    mail_host = mail_host or 'smtp.163.com'
    mail_user = mail_user or 'bert4torch'
    mail_pwd = mail_pwd or 'VDSGQEHFXDZOCVEH'
    mail_sender = mail_sender or 'bert4torch@163.com'

    #构造邮件内容
    message = MIMEText(mail_msg,'plain','utf-8')
    message['Subject'] = mail_subject
    message['From'] = mail_sender
    assert isinstance(mail_receivers, (str, tuple, list)), 'Arg `receivers` only receive `str, tuple, list` format'
    message['To'] = mail_receivers if isinstance(mail_receivers, str) else ';'.join(mail_receivers)

    #登录并发送邮件
    try:
        smtpObj = smtplib.SMTP() 
        smtpObj.connect(mail_host, 25)  # 连接到服务器
        smtpObj.login(mail_user, mail_pwd)  # 登录到服务器
        smtpObj.sendmail(mail_sender, mail_receivers, message.as_string())  # 发送
        smtpObj.quit()  # 退出
        log_info('Send email success')
    except smtplib.SMTPException as e:
        log_error('Send email error : '+str(e))
        return str(e)


def email_when_error(receivers:Union[str,list], **configs):
    '''装饰器, 异常则发邮件
    Examples:
    ```python
    >>> @email_when_error(receivers='tongjilibo@163.com')
    >>> def test():
    ...     return 1/0
    >>> test()  # 调用
    ```
    '''
    def actual_decorator(func):
        def new_func(*args, **kwargs):
            try:
                res = func(*args, **kwargs)
            except Exception as e:
                error_msg = traceback.format_exc()
                send_email(receivers, func.__name__, error_msg, **configs)
                raise e
            return res
        return new_func
    return actual_decorator


def watch_system_state(log_dir:str, gpu_id_list:List[int]=None, pids:Union[int,List[int]]=None, interval=1):
    '''监控system的状态
    
    :param log_dir: str, tensorboard的地址
    :param gpu_id_list: List[int], 监控的gpu
    :param pids: int/List[int], 监控的进程号

    Examples:
    ```python
    >>> watch_system_state(log_dir='./system_states')
    ```
    '''
    import psutil
    import pynvml
    from tensorboardX import SummaryWriter

    pynvml.nvmlInit()
    os.makedirs(log_dir, exist_ok=True)
    tb_writer = SummaryWriter(log_dir=str(log_dir))  # prepare summary writer

    pre_time_int = int(time.time())
    pre_time = time.time()
    pids = pids or os.getpid()
    pids = [pids] if isinstance(pids, int) else pids
    G = 1024*1024*1024

    pre_read = {pid:psutil.Process(pid).io_counters().read_bytes for pid in pids}
    pre_write = {pid:psutil.Process(pid).io_counters().write_bytes for pid in pids}
    time_str_init = int(time.time())

    log_info("Watching System Info")
    while True:
        time.sleep(interval)
        now_time = time.time()
        time_str = int(now_time) - time_str_init
        for pid in pids:
            p = psutil.Process(pid)

            # CPU使用情况
            tb_writer.add_scalar(f"Pid_{pid}/CPU Percent", p.cpu_percent(interval=0.5), time_str)

            # 内存使用情况
            tb_writer.add_scalar(f"Pid_{pid}/Memory Percent", p.memory_percent(), time_str)
            tb_writer.add_scalar(f"Pid_{pid}/Memory RSS G_byte", p.memory_info().rss/G, time_str)
            tb_writer.add_scalar(f"Pid_{pid}/Memory VMS G_byte", p.memory_info().vms/G, time_str)

            # 进程IO信息
            data = p.io_counters()
            tb_writer.add_scalar(f"Pid_{pid}/IO read M_byte", data.read_bytes/1024, time_str)
            tb_writer.add_scalar(f"Pid_{pid}/IO write M_byte", data.write_bytes/1024, time_str)
            if (time_str - pre_time_int)%5 == 0:
                tb_writer.add_scalar(f"Pid_{pid}/IO readRate M_byte_s", (data.read_bytes - pre_read[pid])/(now_time-pre_time)/1024, time_str)
                tb_writer.add_scalar(f"Pid_{pid}/IO writeRate M_byte_s", (data.write_bytes - pre_write[pid])/(now_time-pre_time)/1024, time_str)
                pre_read[pid] = data.read_bytes
                pre_write[pid] = data.write_bytes
            pre_time = now_time
        
        # gpu使用情况
        tesorboard_info_list = []
        if gpu_id_list is None:
            deviceCount = pynvml.nvmlDeviceGetCount()
            device_list = [i for i in range(deviceCount)]
        else:
            device_list = gpu_id_list

        tesorboard_info_list = []
        for i in device_list:
            tesorboard_info = {}
            handle = pynvml.nvmlDeviceGetHandleByIndex(i)
            tesorboard_info['gpu_name'] = pynvml.nvmlDeviceGetName(handle)
            meminfo = pynvml.nvmlDeviceGetMemoryInfo(handle)
            tesorboard_info['gpu_mem_used'] = meminfo.used/G
            UTIL = pynvml.nvmlDeviceGetUtilizationRates(handle)
            tesorboard_info['gpu_gpu_util'] = UTIL.gpu
            tesorboard_info['gpu_mem_util'] = UTIL.memory
            tesorboard_info_list.append(copy.deepcopy(tesorboard_info))
        for tesorboard_info,i in zip(tesorboard_info_list, device_list):
            tb_writer.add_scalar(f'GPU/GPU{i} mem_used unit_G', tesorboard_info['gpu_mem_used'], time_str)
            tb_writer.add_scalar(f'GPU/GPU{i} gpu_util', tesorboard_info['gpu_gpu_util'], time_str)
            tb_writer.add_scalar(f'GPU/GPU{i} mem_util', tesorboard_info['gpu_mem_util'], time_str)

    tb_writer.close()
    pynvml.nvmlShutdown()

    return



from torch import nn
from torch4keras.snippets import log_info, log_warn
from .base import Trainer
from .dp import DPTrainer
from .ddp import DDPTrainer


def add_trainer(obj, include=None, exclude=None, verbose=0, replace_func=False):
    '''为nn.Module添加Triner对应的方法'''
    if isinstance(obj, (Trainer, DPTrainer, DDPTrainer)):
        log_warn('obj is not a Trainer object')
        return obj
    elif not isinstance(obj, nn.Module):
        log_warn('obj is not a nn.Module object')
        return obj

    if include is None:
        include = set()
    elif isinstance(include, str):
        include = set([include])
    elif isinstance(include, (tuple, list)):
        include = set(include)
    else:
        raise TypeError(f'Arg `include` only receive str/list format, not {type(include)}')

    if exclude is None:
        exclude = set()
    elif isinstance(exclude, (tuple, list)):
        exclude = set(exclude)

    import types
    added_funcs = []
    for k in dir(Trainer):
        set_func = False
        if k in include:  # 必须包含的
            set_func = True
        elif k in exclude:  # 必须屏蔽的
            pass
        elif k.startswith('__') and k.endswith('__'):  # 内部函数不执行
            pass
        elif hasattr(obj, k):  # 如果下游重新定义, 则不继
            if replace_func:
                set_func = True
        else:
            set_func = True

        if set_func and eval(f'isfunction(Trainer.{k})'):
            exec(f'obj.{k} = types.MethodType(Trainer.{k}, obj)')
            added_funcs.append(k)
    obj.initialize()  # 这里初始化会得到一些其他的成员变量, 不可缺省

    if verbose and (len(added_funcs) > 0):
        log_info(f'Already add `{",".join(added_funcs)}` method')
    return obj


def add_module(obj, include=None, exclude=None, verbose=0, replace_func=False):
    '''为Trainer增加nn.Module的方法
    方便外部访问, 如obj.parameters()可以直接访问到obj.module.parameters()
    '''
    if isinstance(obj, nn.Module):
        return obj
    elif not isinstance(obj, Trainer):
        log_warn('obj is not a Trainer object')
        return obj
    elif not isinstance(obj.unwrap_model(), nn.Module):
        log_warn('obj.unwrap_model() is not a nn.Module object')
        return obj
    
    if include is None:
        include = set()
    elif isinstance(include, str):
        include = set([include])
    elif isinstance(include, (tuple, list)):
        include = set(include)
    else:
        raise TypeError(f'Arg `include` only receive str/list format, not {type(include)}')

    if exclude is None:
        exclude = set()
    elif isinstance(exclude, (tuple, list)):
        exclude = set(exclude)


    import types
    added_funcs = []
    for k in dir(obj.unwrap_model()):
        set_func = False
        if k in include:  # 必须包含的
            set_func = True
        elif k in exclude:  # 必须屏蔽的
            pass
        elif k.startswith('__') and k.endswith('__'):
            pass
        elif hasattr(obj, k):  # 如果下游重新定义, 则不继
            if replace_func:
                set_func = True
        else:
            set_func = True
        
        if set_func and eval(f'isinstance(obj.unwrap_model().{k}, types.MethodType)'):
            exec(f'obj.{k} = obj.unwrap_model().{k}')
            added_funcs.append(k)

    if verbose and (len(added_funcs) > 0):
        log_info(f'Already add `{",".join(added_funcs)}` method')
    return obj



import numpy as np
from typing import Any, Callable, Dict, List, Optional, Tuple, Union
from packaging import version
import inspect
from .import_utils import is_safetensors_available, is_sklearn_available, is_torch_available
import os


if is_torch_available():
    import torch
    from torch.utils.data import Dataset, IterableDataset
    from torch import nn, Tensor
    from torch.nn import Module
else:
    class Tensor: pass
    class Module: pass
    class Dataset: pass
    class IterableDataset: pass


if is_safetensors_available():
    from safetensors import safe_open
    from safetensors.torch import load_file as safe_load_file
    from safetensors.torch import save_file as safe_save_file


if is_sklearn_available():
    from sklearn.metrics import roc_auc_score
else:
    roc_auc_score = None


def take_along_dim(input_tensor:Tensor, indices:Tensor, dim:int=None):
    '''兼容部分低版本pytorch没有torch.take_along_dim
    '''
    if version.parse(torch.__version__) > version.parse('1.8.1'):
        return torch.take_along_dim(input_tensor, indices, dim)
    else:
        # 该逻辑仅在少量数据上测试, 如有bug, 欢迎反馈
        if dim is None:
            res = input_tensor.flatten()[indices]
        else:
            res = np.take_along_axis(input_tensor.cpu().numpy(), indices.cpu().numpy(), axis=dim)
            res = torch.from_numpy(res).to(input_tensor.device)
        # assert res.equal(torch.take_along_dim(input_tensor, indices, dim))
        return res


def torch_div(input:Tensor, other:Tensor, rounding_mode:Optional[str] = None):
    ''' torch.div兼容老版本
    '''
    if version.parse(torch.__version__) < version.parse('1.7.2'):
        indices = input // other  # 兼容老版本
    else:
        indices = torch.div(input, other, rounding_mode=rounding_mode)  # 行索引
    return indices


def softmax(x:np.ndarray, axis:int=-1):
    '''numpy版softmax
    '''
    x = x - x.max(axis=axis, keepdims=True)
    x = np.exp(x)
    return x / x.sum(axis=axis, keepdims=True)


def search_layer(model:Module, layer_name:str, retrun_first:bool=True):
    '''根据layer_name搜索并返回参数/参数list
    '''
    return_list = []
    for name, param in model.named_parameters():
        if param.requires_grad and layer_name in name:
            return_list.append(param)
    if len(return_list) == 0:
        return None
    if retrun_first:
        return return_list[0]
    else:
        return return_list


class ListDataset(Dataset):
    '''数据是List格式Dataset, 支持传入file_path或者外部已读入的data(List格式)

    :param file_path: str, 待读取的文件的路径, 若无可以为None
    :param data: List[Any], list格式的数据, 和file_path至少有一个不为None
    '''
    def __init__(self, file_path:Union[str, tuple, list]=None, data:Optional[list]=None, **kwargs):
        self.kwargs = kwargs
        if isinstance(file_path, (str, tuple, list)):
            self.data = self.load_data(file_path)
        elif isinstance(data, list):
            self.data = data
        else:
            raise ValueError('The input args shall be str format file_path / list format dataset')

    def __len__(self):
        return len(self.data)

    def __getitem__(self, index):
        return self.data[index]

    @staticmethod
    def load_data(file_path):
        return file_path


class IterDataset(IterableDataset):
    '''流式读取文件, 用于大数据量、多小文件使用时候需要注意steps_per_epoch != None

    :param file_path: str, 待读取的文件的路径, 若无可以为None
    '''
    def __init__(self, file_path:Union[str, tuple, list]=None, **kwargs):
        self.kwargs = kwargs
        if isinstance(file_path, (str, tuple, list)):
            self.file_path = file_path
        else:
            raise ValueError('The input args shall be str format file_path / list format dataset')
    
    def __iter__(self):
        return self.load_data(self.file_path)

    @staticmethod
    def load_data(file_path:Union[str,tuple,list], verbose=0):
        if isinstance(file_path, (tuple, list)):
            for file in file_path:
                if verbose != 0:
                    print("Load data: ", file)
                with open(file, 'r') as file_obj:
                    for line in file_obj:
                        yield line
        elif isinstance(file_path, str):
            if verbose != 0:
                print("Load data: ", file_path)
            with open(file_path, 'r') as file_obj:
                for line in file_obj:
                    yield line


def set_precision(num:float, dense_round:int=1):
    '''设置数字的精度
    '''
    if np.isinf(num):
        return num
    if abs(num) >= 10:
        return int(round(num))

    precision = [1, 0.1, 0.01, 0.001, 0.0001, 0.00001, 0.000001]
    for len_, prec in enumerate(precision):
        if abs(num) >= prec:
            return round(num, len_ + dense_round)
    return num


def metric_mapping(metric:str, func:Callable, y_pred:Union[Tensor, List[Tensor], Tuple[Tensor]], 
                   y_true:Union[Tensor, List[Tensor], Tuple[Tensor]]):
    '''metric的计算

    :param metric: str, 自带metrics的名称
    :param func: function, 透传的用户自定的计算指标的函数
    :param y_pred: Tensor, 样本的预测结果
    :param y_true: Tensor, 样本的真实结果
    '''
    # 自定义metrics
    if inspect.isfunction(func):
        metric_res = func(y_pred, y_true)
        if inspect.isfunction(metric):
            # 如果直接传入回调函数（无key）, 要求回调函数返回Dict[String: Int/Float]类型
            assert isinstance(metric_res, dict), 'Custom metrics callbacks should return `Dict[String: Int/Float]` value'
        elif isinstance(metric, str):
            # 如果直接传入回调函数（有key）, 要求回调函数返回Int/Float类型
            assert isinstance(metric_res, (int, float)), 'Custom metrics callbacks should return `Int/Float value'
        return metric_res
    elif metric == 'loss':
        pass
    # 自带metrics
    elif isinstance(metric, str):
        # 如果forward返回了list, tuple, 则选取第一项
        y_pred_tmp = y_pred[0] if isinstance(y_pred, (list, tuple)) else y_pred
        y_true_tmp = y_true[0] if isinstance(y_true, (list, tuple)) else y_true
        y_pred_tmp = y_pred_tmp.detach()  # 训练过程中评估, detach不进入计算图

        # 根据shape做预处理
        if len(y_pred_tmp.shape) == len(y_true_tmp.shape) + 1:
            y_pred_tmp = torch.argmax(y_pred_tmp, dim=-1)
        elif len(y_pred_tmp.shape) == len(y_true_tmp.shape):
            pass
        else:
            raise ValueError(f'y_pred_tmp.shape={y_pred_tmp.shape} while y_true_tmp.shape={y_true_tmp.shape}')

        # 执行内置的metric
        if metric in {'accuracy', 'acc'}:
            return torch.sum(y_pred_tmp.eq(y_true_tmp)).item() / y_true_tmp.numel()
        elif metric in {'auc'}:
            if roc_auc_score is None:
                raise ImportError('roc_auc_score requires the `sklearn` library.')
            return roc_auc_score(y_true.cpu().numpy(), y_pred_tmp.cpu().numpy())            
        elif metric in {'mae', 'MAE', 'mean_absolute_error'}:
            return torch.mean(torch.abs(y_pred_tmp - y_true_tmp)).item()
        elif metric in {'mse', 'MSE', 'mean_squared_error'}:
            return torch.mean(torch.square(y_pred_tmp - y_true_tmp)).item()
        elif metric in {'mape', 'MAPE', 'mean_absolute_percentage_error'}:
            diff = torch.abs((y_true_tmp - y_pred_tmp) / torch.clamp(torch.abs(y_true_tmp), 1e-7, None))
            return 100. * torch.mean(diff).item()
        elif metric in {'msle', 'MSLE', 'mean_squared_logarithmic_error'}:
            first_log = torch.log(torch.clamp(y_pred_tmp, 1e-7, None) + 1.)
            second_log = torch.log(torch.clamp(y_true_tmp, 1e-7, None) + 1.)
            return torch.mean(torch.square(first_log - second_log)).item()

    return None


def load_checkpoint(checkpoint:str, load_safetensors:bool=False):
    '''加载ckpt, 支持torch.load和safetensors
    '''
    if load_safetensors or checkpoint.endswith(".safetensors"):
        # 加载safetensors格式
        with safe_open(checkpoint, framework="pt") as f:
            metadata = f.metadata()
        if metadata.get("format") not in ["pt", "tf", "flax"]:
            raise OSError(
                f"The safetensors archive passed at {checkpoint} does not contain the valid metadata. Make sure "
                "you save your model with the `save_pretrained` method."
            )
        elif metadata["format"] != "pt":
            raise NotImplementedError(
                f"Conversion from a {metadata['format']} safetensors archive to PyTorch is not implemented yet."
            )
        return safe_load_file(checkpoint)
    else:
        # 正常加载pytorch_model.bin
        return torch.load(checkpoint, map_location='cpu')


def save_checkpoint(state_dict:dict, save_path:str, save_safetensors:bool=False):
    '''保存ckpt, 支持torch.save和safetensors
    '''
    save_dir = os.path.dirname(save_path)
    if save_dir:
        os.makedirs(save_dir, exist_ok=True)

    if save_safetensors or save_path.endswith('.safetensors'):
        safe_save_file(state_dict, save_path, metadata={"format": "pt"})
    else:
        torch.save(state_dict, save_path)


from torch.utils.data import DataLoader
from torch import nn
from torch4keras.snippets import JsonConfig, log_info, log_warn, log_warn_once
from torch4keras.snippets import argument_parse
from torch4keras.snippets import print_table, json_flat
from typing import Union
import os
import math
import inspect
from .base import Trainer


class DeepSpeedTrainer(Trainer):
    '''deepspeed来训练'''
    def __init__(self, module:nn.Module, verbose:int=1, **kwargs):
        super().__init__(module)
        self.module = module

        # 参数解析
        ds_args = DeepSpeedArgs(argument_parse())  # 解析命令行参数
        ds_args.set_default_args()  # 设置默认的一些参数
        ds_args.trainer_config_process(self.module, auto_find_batch_size=False)  # 设置一些auto的参数
        self.ds_config = ds_args.ds_config

        if verbose > 0:
            log_info('Deepspeed config listed below.')
            print_table(json_flat(self.ds_config), headers=['ds_config name', 'ds_config value'])
    
    def _prepare_inputs(self, train_dataloader:DataLoader, steps_per_epoch:Union[int,None], epochs:int, verbose:int):
        # batch_size需要使用deepspeed config中的train_batch_size/train_micro_batch_size_per_gpu
        if train_dataloader.batch_sampler is not None:
            btz = train_dataloader.batch_sampler.batch_size
            btz_ds = self.ds_config.train_batch_size
            btz_ds_per = self.ds_config.train_micro_batch_size_per_gpu
            if btz != btz_ds:
                log_warn_once(f'Use deepspeed config `train_batch_size`={btz_ds} and `train_micro_batch_size_per_gpu`={btz_ds_per} instead of `batch_size`={btz}')
            train_dataloader.batch_sampler.batch_size = self.ds_config.train_batch_size
        super()._prepare_inputs(train_dataloader, steps_per_epoch, epochs, verbose)


    def compile(self, *args, log_level='warning', inference=False, master_rank=0, **kwargs):
        super().compile(*args, **kwargs)
        import deepspeed
        from deepspeed.utils import logger as ds_logger
        import logging
        log_levels = {
            "debug": logging.DEBUG,
            "info": logging.INFO,
            "warning": logging.WARNING,
            "error": logging.ERROR,
            "critical": logging.CRITICAL,
        }

        ds_logger.setLevel(log_levels.get(log_level, logging.WARNING))

        if inference:
            # only Z3 makes sense for the inference
            log_warn("ZeRO inference only makes sense with ZeRO Stage 3")
            self.optimizer, self.scheduler = None, None
            model_parameters = None
        else:
            model_parameters = list(filter(lambda p: p.requires_grad, self.module.parameters()))
        
        kwargs = {
            "model": self.module,  # deepspeed的forward默认是计算到loss输出的
            "model_parameters": model_parameters,
            "config_params": self.ds_config,
            "optimizer": self.optimizer,
            "lr_scheduler": self.scheduler,
        }
        if self.ds_config.get('zero_optimization', {}).get('offload_optimizer', {}).get('device') == 'cpu':
            kwargs.pop('optimizer')
            if self.optimizer is not None:
                self.optimizer = None
                log_warn('You may not use custom optimizer when offload_optimizer=`cpu`')
        self.deepspeed_engine, self.optimizer, _, self.scheduler = deepspeed.initialize(**kwargs)
        self.verbose = 1 if self.deepspeed_engine.local_rank == master_rank else 0

    def unwrap_model(self) -> nn.Module:
        # 执行deepspeed_engine的forward
        return self.deepspeed_engine

    def loss_backward(self, loss):
        self.deepspeed_engine.backward(loss)
        return loss
    
    def step(self):
        self.deepspeed_engine.step()

    def resume_from_checkpoint(self, *args, verbose:int=1, **kwargs):
        from deepspeed import DeepSpeedEngine
        kwargs_ = {
            k: v for k, v in kwargs.items() if k in inspect.signature(DeepSpeedEngine.load_checkpoint).parameters
        }
        save_dir = args[0] if len(args) > 0 else kwargs['save_dir']
        self.deepspeed_engine.load_checkpoint(save_dir, **kwargs_)
        self.load_steps_params(os.path.join(save_dir, 'steps_params.pt'))
        if verbose == 1:
            log_info('Successfuly resume training checkpoint')

    def save_to_checkpoint(self, *args, verbose:int=0, **kwargs):
        from deepspeed import DeepSpeedEngine
        kwargs_ = {
            k: v for k, v in kwargs.items() if k in inspect.signature(DeepSpeedEngine.save_checkpoint).parameters
        }
        save_dir = args[0] if len(args) > 0 else kwargs['save_dir']
        self.deepspeed_engine.save_checkpoint(save_dir, **kwargs_)
        self.save_steps_params(os.path.join(save_dir, 'steps_params.pt'))
        if verbose == 1:
            log_info('Successfuly save training checkpoint')


class DeepSpeedArgs():
    '''deepspeed的config设置, 含自动填充auto
    为了和transformers保持一致，部分命令行参数名尽量和其保持一致，好处是一些启动命令可以直接套用
    '''
    def __init__(self, arguements) -> None:
        self.train_args = arguements
        self.ds_config = JsonConfig(arguements.deepspeed)
        
    def set_default_args(self):
        '''设置默认的参数，用于deepspeed里面参数设置为auto的情况'''
        self.ds_config.steps_per_print = self.ds_config.get('steps_per_print', 1e9)  # 默认不打印, 防止进度条打印问题

        self.default_args = set()
        def set_default_arg(name, value):
            # 命令行参数未传入的，设置为默认值
            if name not in self.train_args:
                self.train_args[name] = value
                self.default_args.add(name)

        # 训练参数的默认参数
        set_default_arg('world_size', int(os.environ["WORLD_SIZE"]))
        set_default_arg('per_device_train_batch_size', 8)
        set_default_arg('gradient_accumulation_steps', 1)
        set_default_arg('max_grad_norm', 1.0)
        set_default_arg('learning_rate', 5e-5)
        set_default_arg('adam_beta1', 0.9)
        set_default_arg('adam_beta2', 0.999)
        set_default_arg('adam_epsilon', 1e-8)
        set_default_arg('weight_decay', 0.0)
        set_default_arg('fp16', False)
        set_default_arg('fp16_full_eval', False)
        set_default_arg('fp16_opt_level', "O1")
        set_default_arg('fp16_backend', "auto")
        set_default_arg('bf16', False)
        set_default_arg('bf16_full_eval', False)
        set_default_arg('warmup_steps', 0)
        set_default_arg('warmup_ratio', 0.0) 

    def find_config_node(self, ds_key_long):
        config = self.ds_config

        # find the config node of interest if it exists
        nodes = ds_key_long.split(".")
        ds_key = nodes.pop()
        for node in nodes:
            config = config.get(node)
            if config is None:
                return None, ds_key

        return config, ds_key
    
    def fill_match(self, ds_key_long, b4t_val, must_match=True):
        """填充auto的默认值, 使用 "命令行参数/默认值"
        """
        config, ds_key = self.find_config_node(ds_key_long)
        if config is None:
            return

        if config.get(ds_key) == "auto":
            config[ds_key] = b4t_val
            return

        if not must_match:
            return

    def get_value(self, ds_key_long, default=None):
        """
        Returns the set value or `default` if no value is set
        """
        config, ds_key = self.find_config_node(ds_key_long)
        if config is None:
            return default
        return config.get(ds_key, default)
    
    def trainer_config_process(self, model, auto_find_batch_size=False):
        """自动填充和替换ds_config中的auto选项
        """
        args = self.train_args

        # DeepSpeed does:
        # train_batch_size = world_size * train_micro_batch_size_per_gpu * gradient_accumulation_steps
        train_batch_size = args.world_size * args.per_device_train_batch_size * args.gradient_accumulation_steps
        self.fill_match("train_micro_batch_size_per_gpu", args.per_device_train_batch_size, must_match=not auto_find_batch_size)
        self.fill_match("gradient_accumulation_steps", args.gradient_accumulation_steps)
        self.fill_match("train_batch_size", train_batch_size, must_match = not auto_find_batch_size)
        self.fill_match("gradient_clipping", args.max_grad_norm)

        self.fill_match("optimizer.params.lr", args.learning_rate)
        self.fill_match("optimizer.params.betas", [args.adam_beta1, args.adam_beta2])
        self.fill_match("optimizer.params.eps", args.adam_epsilon)
        self.fill_match("optimizer.params.weight_decay", args.weight_decay)

        self.fill_match("scheduler.params.warmup_min_lr", 0, must_match=False)  # not a trainer arg
        self.fill_match("scheduler.params.warmup_max_lr", args.learning_rate)
        # total_num_steps - will get set in trainer_config_finalize

        # fp16
        if args.fp16 or args.fp16_full_eval:
            fp16_backend = "apex" if args.fp16_backend == "apex" else "amp"
        else:
            fp16_backend = None

        # amp: similar to the pytorch native amp - it has a bunch of optional params but we won't set
        # any here unless the user did the work
        self.fill_match("fp16.enabled", ((args.fp16 or args.fp16_full_eval) and fp16_backend == "amp"))

        # apex: delegates amp work to apex (which needs to be available), but it cannot be used with any
        # ZeRO features
        self.fill_match("amp.enabled", fp16_backend == "apex")
        self.fill_match("amp.opt_level", args.fp16_opt_level)

        self.fill_match("bf16.enabled", (args.bf16 or args.bf16_full_eval))

        ''' 以下逻辑为transformers中trainer_config_finalize修改'''
        # deal with config keys that use `auto` value and rely on model's hidden_size
        hidden_size_based_keys = [
            "zero_optimization.reduce_bucket_size",
            "zero_optimization.stage3_prefetch_bucket_size",
            "zero_optimization.stage3_param_persistence_threshold",
        ]
        hidden_size_auto_keys = [x for x in hidden_size_based_keys if self.get_value(x) == 'auto']

        if len(hidden_size_auto_keys) > 0:
            if hasattr(model.config, "hidden_size"):
                hidden_size = self.ds_config.hidden_size
            elif hasattr(self.ds_config, "hidden_sizes"):
                # if there are many hidden sizes pick the largest one
                hidden_size = max(self.ds_config.hidden_sizes)
            else:
                raise ValueError(
                    "The model's config file has neither `hidden_size` nor `hidden_sizes` entry, "
                    "therefore it's not possible to automatically fill out the following `auto` entries "
                    f"in the DeepSpeed config file: {hidden_size_auto_keys}. You can fix that by replacing "
                    "`auto` values for these keys with an integer value of your choice."
                )

            self.fill_match("zero_optimization.reduce_bucket_size", hidden_size * hidden_size, must_match=False)
            if self.get_value("zero_optimization.stage", -1) == 3:
                # automatically assign the optimal config values based on model config
                self.fill_match("zero_optimization.stage3_prefetch_bucket_size", 0.9 * hidden_size * hidden_size, must_match=False)
                self.fill_match("zero_optimization.stage3_param_persistence_threshold", 10 * hidden_size, must_match=False)

        # scheduler
        if hasattr(self, 'totel_steps'):
            self.fill_match("scheduler.params.total_num_steps", self.total_steps)
            self.fill_match("scheduler.params.warmup_num_steps", self.ds_config.warmup_steps if self.ds_config.warmup_steps > 0 
                             else math.ceil(self.total_steps * self.ds_config.warmup_ratio))


from torch import nn
import torch
from torch.utils.data import DataLoader
from torch4keras.snippets import DottableDict, seed_everything
from torch4keras.callbacks import Callback
from typing import Union, List
import os
from .base import Trainer


class DDPTrainer(nn.parallel.DistributedDataParallel, Trainer):
    '''DistributedDataParallel模式使用多gpu的方法,
    1) 父类顺序颠倒也会出问题
    2) 使用方式和DistributedDataParallel一致, DDPTrainer(net, *args, **kwargs)来使用

    Examples:
    ```python
    >>> from torch4keras.trainer import DDPTrainer
    >>> args = DDPTrainer.init_process_group()
    >>> # 用户自行定义好model
    >>> model = DDPTrainer(
    ...     model,
    ...     master_rank=0,
    ...     device_ids=[args.local_rank],
    ...     output_device=args.local_rank,
    ...     find_unused_parameters=False
    ... )
    >>> # 后续执行和Trainer一致
    ```
    '''
    def __init__(self, *args, master_rank=0, **kwargs):
        Trainer.__init__(self)
        kwargs['device_ids'] = kwargs.get('device_ids', [int(os.getenv('LOCAL_RANK'))])
        kwargs['output_device'] = kwargs.get('output_device', int(os.getenv('LOCAL_RANK')))
        nn.parallel.DistributedDataParallel.__init__(self, *args, **kwargs)

        # 默认仅对master_rank=0打印信息
        assert isinstance(master_rank, (int, list, tuple)), 'Args `master_rank` only supoorts int, list, tuple'
        if isinstance(master_rank, int):
            master_rank = [master_rank]
        self.master_rank = master_rank
        self.verbose = (torch.distributed.get_rank() in master_rank)
    
    def _prepare_inputs(self, train_dataloader:DataLoader, steps_per_epoch:Union[int,None], epochs:int, verbose:int):
        # 如果使用ddp的时候没有使用DistributedSampler，这里会自动修改一下
        from torch.utils.data.distributed import DistributedSampler 
        if (train_dataloader.sampler is None) and (not isinstance(train_dataloader.sampler, DistributedSampler)):
            train_dataloader.sampler = DistributedSampler(train_dataloader.dataset)
        super()._prepare_inputs(train_dataloader, steps_per_epoch, epochs, verbose)
    
    def disable_workers_callback(self, callbacks: Union[Callback, List[Callback]]):
        '''非master_rank上不使用callback'''
        for callback in callbacks:
            if torch.distributed.get_rank() not in self.master_rank:
                callback.run_callback = False

    @classmethod
    def init_process_group(cls, master_rank=0, seed=42):
        '''初始化各项参数'''
        cls.ddp_config = DottableDict()
        cls.ddp_config.rank = int(os.environ["RANK"])
        cls.ddp_config.local_rank = int(os.getenv('LOCAL_RANK'))
        cls.ddp_config.device = torch.device('cuda', cls.ddp_config.local_rank)
        cls.ddp_config.world_size = int(os.environ["WORLD_SIZE"])
        cls.ddp_config.master_process = cls.ddp_config.rank == master_rank
        torch.cuda.set_device(cls.ddp_config.local_rank)
        seed_everything(seed + cls.ddp_config.rank)
        return cls.ddp_config



import numpy as np
from typing import Any, Callable, Dict, List, Optional, Tuple, Union
import os
import random
from .log import log_info, log_warn, log_error, log_warn_once, print_table
from .monitor import format_timestamp, format_time
from .import_utils import is_torch_available
import json
import time
import sys
import shutil
import re


if is_torch_available():
    import torch
    from torch import Tensor
    from torch.nn import Module
else:
    class Tensor: pass
    class Module: pass


def seed_everything(seed:int=None):
    '''固定seed
    
    :param seed: int, 随机种子
    '''
    max_seed_value = np.iinfo(np.uint32).max
    min_seed_value = np.iinfo(np.uint32).min

    if (seed is None) or not (min_seed_value <= seed <= max_seed_value):
        seed = random.randint(np.iinfo(np.uint32).min, np.iinfo(np.uint32).max)
    log_info(f"Global seed set to {seed}")
    os.environ["PYTHONHASHSEED"] = str(seed)
    random.seed(seed)
    np.random.seed(seed)
    torch.manual_seed(seed)
    torch.cuda.manual_seed(seed)
    torch.cuda.manual_seed_all(seed)
    torch.backends.cudnn.benchmark = False
    torch.backends.cudnn.deterministic = True
    return seed


def print_trainable_parameters(module:Module):
    '''打印可训练的参数量'''
    trainable_params = 0
    all_param = 0
    for _, param in module.named_parameters():
        num_params = param.numel()
        # if using DS Zero 3 and the weights are initialized empty
        if num_params == 0 and hasattr(param, "ds_numel"):
            num_params = param.ds_numel

        all_param += num_params
        if param.requires_grad:
            trainable_params += num_params
    log_info(f"trainable params: {trainable_params} || all params: {all_param} || trainable%: {100 * trainable_params / all_param}")


def get_parameter_device(parameter):
    '''获取device, 从transformers包迁移过来'''
    try:
        return next(parameter.parameters()).device
    except StopIteration:
        # For nn.DataParallel compatibility in PyTorch 1.5

        def find_tensor_attributes(module: Module) -> List[Tuple[str, Tensor]]:
            tuples = [(k, v) for k, v in module.__dict__.items() if torch.is_tensor(v)]
            return tuples

        gen = parameter._named_members(get_members_fn=find_tensor_attributes)
        first_tuple = next(gen)
        return first_tuple[1].device


class DottableDict(dict):
    '''支持点操作符的字典'''
    def __init__(self, *args, **kwargs):
        dict.__init__(self, *args, **kwargs)
        self.__dict__ = self
    def allowDotting(self, state=True):
        if state:
            self.__dict__ = self
        else:
            self.__dict__ = dict()
KwargsConfig = DottableDict


def tran2dottableDict(dict_:dict):
    '''将一个嵌套字典转成DottableDict'''
    def traverse_dict(d): 
        d = DottableDict(d)
        for key, value in d.items():
            if isinstance(value, dict):
                d[key] = traverse_dict(value)
        return d
    return traverse_dict(dict_)


class JsonConfig(DottableDict):
    '''读取json配置文件并返回可.操作符的字典'''
    def __new__(cls, json_path:str, encoding:str='utf-8', dot:bool=True):
        data = json.load(open(json_path, "r", encoding=encoding))
        if dot:
            return tran2dottableDict(data)
        else:
            return data


class YamlConfig(DottableDict):
    '''读取yaml配置文件并返回可.操作符的字典'''
    def __new__(cls, yaml_path:str, encoding:str='utf-8', dot:bool=True):
        import yaml
        data = yaml.load(open(yaml_path, "r", encoding=encoding), Loader=yaml.FullLoader)
        if dot:
            return tran2dottableDict(data)
        else:
            return data

class IniConfig(DottableDict):
    '''读取ini配置文件'''
    def __new__(cls, ini_path:str, encoding:str='utf-8', dot:bool=True):
        import configparser
        config = configparser.ConfigParser()
        config.read(ini_path, encoding=encoding)
        if dot:
            return DottableDict({section: DottableDict(config.items(section)) for section in config.sections()})
        else:
            return config


def auto_set_cuda_devices(best_num: Optional[int] = None) -> str:
    '''
    这段代码是一个名为 auto_set_cuda_devices 的函数, 它接受一个可选的整数参数 best_num。该函数用于自动设置环境变量 CUDA_VISIBLE_DEVICES, 以便在多个 GPU 设备中选择最佳的 GPU 设备。
    首先, 该函数检查环境变量 CUDA_VISIBLE_DEVICES 是否已经设置。如果已经设置, 则发出警告并返回当前设置的值。
    如果 best_num 等于 -1, 则将环境变量 CUDA_VISIBLE_DEVICES 设置为 -1 并返回。
    接下来, 该函数尝试使用 nvidia-smi 命令查询 GPU 设备的信息。如果命令不存在, 则发出警告并将环境变量 CUDA_VISIBLE_DEVICES 设置为 -1 并返回。
    如果未指定 best_num, 则从环境变量 LOCAL_WORLD_SIZE 中获取值。然后将其转换为整数。
    接下来, 该函数解析 nvidia-smi 命令的输出, 并计算每个 GPU 设备的得分。得分由 GPU 利用率和可用内存计算得出。
    最后, 该函数选择得分最高的 best_num 个 GPU 设备, 并将其索引作为字符串连接起来。然后将环境变量 CUDA_VISIBLE_DEVICES 设置为该字符串并返回。
    '''
    import subprocess

    # 无需重复设置
    if "CUDA_VISIBLE_DEVICES" in os.environ:
        log_warn("Environment variable `CUDA_VISIBLE_DEVICES` has already been set")
        return os.environ["CUDA_VISIBLE_DEVICES"]

    if best_num == -1:
        log_info(f"SET CUDA_VISIBLE_DEVICES=-1")
        os.environ["CUDA_VISIBLE_DEVICES"] = "-1"
        return "-1"

    try:
        p = subprocess.Popen(
            ["nvidia-smi",
             "--query-gpu=index,utilization.gpu,memory.total,memory.free",
             "--format=csv,noheader,nounits"],
            stdout=subprocess.PIPE)
    except FileNotFoundError:
        log_error("`nvidia-smi` not exists")
        os.environ["CUDA_VISIBLE_DEVICES"] = "-1"
        return "-1"

    if best_num is None:
        best_num = os.environ.get("LOCAL_WORLD_SIZE", 1)
    best_num = int(best_num)

    stdout, _ = p.communicate()
    outputs = stdout.decode("utf-8")
    lines = outputs.split(os.linesep)

    scores = []
    for item in lines:
        item_arr = item.split(",")
        if len(item_arr) != 4:
            continue
        gpu_score = 100 - int(item_arr[1].strip())
        memory_score = 100 * (int(item_arr[3].strip()) / int(item_arr[2].strip()))
        score = gpu_score + memory_score
        scores.append(score)

    best_num = min(best_num, len(scores))
    topk_idx = (-np.asarray(scores)).argsort()[:best_num]

    topk_idx_str = ",".join(map(str, topk_idx))
    log_info(f"SET CUDA_VISIBLE_DEVICES={topk_idx_str}")
    os.environ["CUDA_VISIBLE_DEVICES"] = topk_idx_str

    return topk_idx_str


def find_tied_parameters(model: Module, **kwargs):
    """ copyed from accelerate
    Find the tied parameters in a given model.

    <Tip warning={true}>

    The signature accepts keyword arguments, but they are for the recursive part of this function and you should ignore
    them.

    </Tip>

    Args:
        model (`torch.Module`): The model to inspect.

    Returns:
        List[List[str]]: A list of lists of parameter names being all tied together.

    Examples:
    ```python
    >>> from collections import OrderedDict
    >>> import torch.nn as nn

    >>> model = nn.Sequential(OrderedDict([("linear1", nn.Linear(4, 4)), ("linear2", nn.Linear(4, 4))]))
    >>> model.linear2.weight = model.linear1.weight
    >>> find_tied_parameters(model)
    ... # [['linear1.weight', 'linear2.weight']]
    ```
    """
    # Initialize result and named_parameters before recursing.
    named_parameters = kwargs.get("named_parameters", None)
    prefix = kwargs.get("prefix", "")
    result = kwargs.get("result", {})

    if named_parameters is None:
        named_parameters = {n: p for n, p in model.named_parameters()}
    else:
        # A tied parameter will not be in the full `named_parameters` seen above but will be in the `named_parameters`
        # of the submodule it belongs to. So while recursing we track the names that are not in the initial
        # `named_parameters`.
        for name, parameter in model.named_parameters():
            full_name = name if prefix == "" else f"{prefix}.{name}"
            if full_name not in named_parameters:
                # When we find one, it has to be one of the existing parameters.
                for new_name, new_param in named_parameters.items():
                    if new_param is parameter:
                        if new_name not in result:
                            result[new_name] = []
                        result[new_name].append(full_name)

    # Once we have treated direct parameters, we move to the child modules.
    for name, child in model.named_children():
        child_name = name if prefix == "" else f"{prefix}.{name}"
        find_tied_parameters(child, named_parameters=named_parameters, prefix=child_name, result=result)

    # return FindTiedParametersResult([sorted([weight] + list(set(tied))) for weight, tied in result.items()])
    return {weight: list(set(tied)) for weight, tied in result.items()}


def check_cuda_verison():
    '''使用torch查看cuda, cudnn的版本'''
    versions = []
    # 查看PyTorch版本
    versions.append([f'torch', torch.__version__])

    # 查看CUDA版本（如果已安装）
    versions.append(["CUDA version", torch.version.cuda if torch.cuda.is_available() else 'Not available'])
    
    # 查看cuDNN版本（如果已安装）
    versions.append(["cuDNN version", torch.backends.cudnn.version() if torch.backends.cudnn.is_available() else 'Not available'])
    
    log_info('Cuda version summary')
    print_table(versions, headers=['name', 'version'])


def check_cuda_capability():
    '''打印各个显卡的算力'''
    # 查看CUDA版本（如果已安装）
    if torch.cuda.is_available():
        gpu_count = torch.cuda.device_count()
        gpu_summay = []
        for device in range(gpu_count):
            gpu_summay.append([device, torch.cuda.get_device_name(device), str(torch.cuda.get_device_capability(device))])
        log_info('CUDA device capability summary')
        print_table(gpu_summay, headers=['device', 'name', 'capability'])
    else:
        log_error('CUDA not available')


  
def check_file_modify_time(file_path:Union[str, List[str]], duration:int=None, verbose=0):
    """  
    判断文件被修改的时间
    :param file_path: 文件路径
    :param duration: 文件的日期如何在duration以内，认为文件刚刚被修改
    :return: 如果文件被修改，返回True；否则返回False
    """
    if isinstance(file_path, str):
        file_path = [file_path]
    
    results = []
    for file_i in file_path:
        # 获取文件的当前修改时间
        file_mtime = os.path.getmtime(file_i)

        # 比较当前修改时间和上次检查的修改时间
        cur_time = time.time()
        res = {'file_name': os.path.basename(file_i)}
        res['file_modify_time'] = format_timestamp(file_mtime)
        if duration is not None and isinstance(duration, int):
            res['current_time'] = format_timestamp(cur_time)
            diff_duration = cur_time - file_mtime
            res['diff_duration'] = format_time(diff_duration)
            res['modified'] = True if diff_duration <= duration else False
        results.append(res)
    
    if verbose > 0:
        print_table(results)

    if (len(results) == 1) and duration is not None:
        return results[0]['modified']
    else:
        return results


def check_file_modified(file_path:Union[str, List[str]], duration:int=1, verbose=0):
    '''判断文件在duration区间是否被修改
    '''
    return check_file_modify_time(file_path, duration, verbose)


def argument_parse(arguments:Union[str, list, dict]=None, description='argument_parse', parse_known_args:bool=True, dot:bool=True):
    ''' 根据传入的参数接受命令行参数，生成argparse.ArgumentParser
    :param arguments: 参数设置，接受str, list, dict输入
    :param description: 描述
    :param parse_known_args: bool, 只解析命令行中认识的参数

    Examples:
    ```python
    >>> args = argument_parse()
    >>> args = argument_parse('deepspeed')
    >>> args = argument_parse(['deepspeed'])
    >>> args = argument_parse({'deepspeed': {'type': str, 'help': 'deepspeed config path'}})
    ```
    '''
    import argparse
    parser = argparse.ArgumentParser(description=description)

    if arguments is None:
        # 不预设，直接解析所有命令行参数
        arguments = [arg for arg in sys.argv[1:] if arg.startswith('-') and ('=' not in arg)]
            
    if isinstance(arguments, str):
        parser.add_argument(f'--{arguments}')
    elif isinstance(arguments, list):
        for argument in arguments:
            if argument.startswith('-') and ('=' not in argument):
                parser.add_argument(f"--{argument.lstrip('-')}")
    elif isinstance(arguments, dict):
        for argument, kwargs in arguments.items():
            parser.add_argument(f'--{argument}',  **kwargs)
    else:
        raise TypeError('Args `arguments` only accepts `str,list,dict` format')

    if parse_known_args:
        args, unknown_args = parser.parse_known_args()  # 允许其他参数不传入
    else:
        args = parser.parse_args()
    
    if dot is True:
        args = tran2dottableDict(vars(args))
    return args


def cuda_empty_cache(device=None):
    '''清理gpu显存'''
    if torch.cuda.is_available():
        if device is None:
            torch.cuda.empty_cache()
            torch.cuda.ipc_collect()
            return
        with torch.cuda.device(device):
            torch.cuda.empty_cache()
            torch.cuda.ipc_collect()
    else:
        log_warn_once('torch.cuda.is_available() = False')


class WebServing(object):
    """简单的Web接口，基于bottlepy简单封装，仅作为临时测试使用，不保证性能。

    Examples:
    ```python
    >>> arguments = {'text': (None, True), 'n': (int, False)}
    >>> web = WebServing(port=8864)
    >>> web.route('/gen_synonyms', gen_synonyms, arguments)
    >>> web.start()
    >>> # 然后访问 http://127.0.0.1:8864/gen_synonyms?text=你好
    ```
    
    ```shell
    >>> # 依赖（如果不用 server='paste' 的话，可以不装paste库）:
    >>> pip install bottle
    >>> pip install paste
    ```
    """
    def __init__(self, host='0.0.0.0', port=8000, server='paste'):

        import bottle

        self.host = host
        self.port = port
        self.server = server
        self.bottle = bottle

    def wraps(self, func, arguments, method='GET'):
        """封装为接口函数

        :param func: 要转换为接口的函数，需要保证输出可以json化，即需要保证 json.dumps(func(inputs)) 能被执行成功；
        :param arguments: 声明func所需参数，其中key为参数名，value[0]为对应的转换函数（接口获取到的参数值都是字符串型），value[1]为该参数是否必须；
        :param method: 'GET'或者'POST'。
        """
        def new_func():
            outputs = {'code': 0, 'desc': u'succeeded', 'data': {}}
            kwargs = {}
            for key, value in arguments.items():
                if method == 'GET':
                    result = self.bottle.request.GET.getunicode(key)
                else:
                    result = self.bottle.request.POST.getunicode(key)
                if result is None:
                    if value[1]:
                        outputs['code'] = 1
                        outputs['desc'] = 'lack of "%s" argument' % key
                        return json.dumps(outputs, ensure_ascii=False)
                else:
                    if value[0] is not None:
                        result = value[0](result)
                    kwargs[key] = result
            try:
                outputs['data'] = func(**kwargs)
            except Exception as e:
                outputs['code'] = 2
                outputs['desc'] = str(e)
            return json.dumps(outputs, ensure_ascii=False)

        return new_func

    def route(self, path, func, arguments, method='GET'):
        """添加接口"""
        func = self.wraps(func, arguments, method)
        self.bottle.route(path, method=method)(func)

    def start(self):
        """启动服务"""
        self.bottle.run(host=self.host, port=self.port, server=self.server)


def copytree(src:str, dst:str, ignore_copy_files:str=None, dirs_exist_ok=False):
    '''从一个文件夹copy到另一个文件夹
    
    :param src: str, copy from src
    :param dst: str, copy to dst
    '''
    def _ignore_copy_files(path, content):
        to_ignore = []
        if ignore_copy_files is None:
            return to_ignore
        
        for file_ in content:
            for pattern in ignore_copy_files:
                if re.search(pattern, file_):
                    to_ignore.append(file_)
        return to_ignore

    if src:
        os.makedirs(src, exist_ok=True)
    shutil.copytree(src, dst, ignore=_ignore_copy_files, dirs_exist_ok=dirs_exist_ok)



from torch4keras.snippets import LoggerHandler
import time
log = LoggerHandler(log_path='./log/log.log', handles=['StreamHandler', 'TimedRotatingFileHandler'],
                    handle_config={'when': 'S', 'interval': 3, 'backupCount': 2, 'encoding': 'utf-8'})

log.debug('debug')
log.info('info')
log.warning('warning')
log.error('error')
log.critical('critical')

for i in range(10):
    log.info('hello '+str(i))
    time.sleep(2)


from .snippets import is_torch_available, log_warn

if is_torch_available():
    from torch import nn

    # torch2.0以后nn.Module自带compile方法
    # 和model.compile()冲突, 这里把nn.Module.compile重命名为compile_torch
    if hasattr(nn.Module, 'compile'):
        nn.Module.compile_torch = nn.Module.compile
        del nn.Module.compile
else:
    log_warn("PyTorch have not been found. Models won't be available and only snippets can be used.")



from torch import nn
from torch4keras.trainer import *


class BaseModel(nn.Module, Trainer):
    '''BaseModel, 使用继承的方式来使用
    
    Examples:
    ```python
    >>> class MyModel(BaseModel):
    ...     def __init__(self, module):
    ...         super().__init__()
    ...         self.model = module
    ...     def forward(self, inputs):
    ...         return self.model(inputs)

    >>> model = MyModel(net)  # 实例化模型对象
    ```
    '''
    def __init__(self, *args, **kwargs):
        nn.Module.__init__(self)
        Trainer.__init__(self, *args, **kwargs)
        
BaseModelDP = DPTrainer  # 使用方式和nn.DataParallel一致
BaseModelDDP = DDPTrainer  # 使用方式和DistributedDataParallel一致


from torch import nn
from .base import Trainer


class DPTrainer(nn.DataParallel, Trainer):
    '''DataParallel模式使用多gpu的方法, 
    1) 父类顺序颠倒也会出问题
    2) 使用方式和nn.DataParallel一致, DPTrainer(net, *args, **kwargs)来使用
    '''
    def __init__(self, *args, **kwargs):
        Trainer.__init__(self)
        nn.DataParallel.__init__(self, *args, **kwargs)


from torch import nn
import torch
from torch.optim import Optimizer
from torch.optim.lr_scheduler import LambdaLR
from torch.utils.data import DataLoader
from torch4keras.snippets import metric_mapping, get_parameter_device, log_info, log_warn, log_warn_once, print_table
from torch4keras.snippets import print_trainable_parameters, colorful, send_email, load_checkpoint, save_checkpoint
from torch4keras.callbacks import KerasProgbar, SmoothMetricsCallback, TqdmProgbar, ProgressBar2Progbar, Callback, CallbackList, History
from collections import OrderedDict
from typing import Union, List, Literal, Tuple, Set, Callable, Optional
from inspect import isfunction
import os
import math
import re
import traceback


class Trainer:
    '''Trainer, 传入Module实例

    :param module: None/nn.Module, nn.Module()的模型实例

    Examples:
    ```python
    >>> import torch
    >>> import torch.nn as nn
    >>> import torch.optim as optim
    >>> import torchvision
    >>> from torch4keras.model import Trainer
    >>> from torch.utils.data import TensorDataset, DataLoader
     
    >>> device = 'cuda' if torch.cuda.is_available() else 'cpu'
     
    >>> # 读取数据
    >>> mnist = torchvision.datasets.MNIST(root='./', download=True)
    >>> x, y = mnist.train_data.unsqueeze(1).to(device).float() / 255.0, mnist.train_labels.to(device)
    >>> train_dataloader = DataLoader(TensorDataset(x, y), batch_size=8)
     
    >>> # 准备模型
    >>> net = torch.nn.Sequential(
    ...             nn.Conv2d(1, 32, kernel_size=3), nn.ReLU(),
    ...             nn.MaxPool2d(2, 2), 
    ...             nn.Conv2d(32, 64, kernel_size=3), nn.ReLU(),
    ...             nn.Flatten(),
    ...             nn.Linear(7744, 10)
    ...         )
     
    >>> model = Trainer(net.to(device))
    >>> model.compile(optimizer=optim.Adam(net.parameters()), loss=nn.CrossEntropyLoss())
    
    >>> # 模型开始训练
    >>> model.fit(train_dataloader, steps_per_epoch=None, epochs=5)
    ```
    '''
    def __init__(self, module:nn.Module=None):
        super(Trainer, self).__init__()
        self.initialize(module)
    
    def initialize(self, module:nn.Module=None):
        # 传入Module实例方式
        if module is not None:
            if not isinstance(module, nn.Module):
                raise TypeError(f'Args `module` only support nn.Module format not {type(module)}')
            self.module = module

        self.global_step, self.local_step, self.total_steps, self.batch_step = 0, 0, 0, 0
        self.epoch, self.steps_per_epoch, self.train_dataloader = 0, None, None
        self.resume_step, self.resume_epoch, self.resume_batch = 0, 0, 0
        self.retain_graph = False  # loss.backward()是否保留计算图
        self.move_to_model_device = True  # 自动把tensor转到model所在的device
        self.log_first_step = False  # 是否打印第一个step的数据
        self.criterion = None  # criterion
        self.optimizer = None  # optimizer
        self.scheduler = None  # scheduler
        self.callbacks = []  # 所有的Callbacks, 如果fit中不传入, 则默认为[progbarlogger, smoothmetrics, history]三项
        self.run_callbacks = True  # 是否运行Callbacks, 目前主要是在DDP模式下运用
        self.loss2metrics = True  # 把loss_detail打印在进度条的metrics上
        # add_module(self)  # 增加nn.Module的成员方法

    def compile(self, loss:Optional[Union[Callable, nn.Module]]=None, optimizer:Optimizer=None, scheduler:LambdaLR=None, clip_grad_norm:float=None, 
                mixed_precision:Literal[True, False, 'fp16', 'bf16']=False, metrics:Union[str, dict, Callable, List[Union[str, dict, Callable]]]=None, 
                grad_accumulation_steps:int=1, progbar_type:Literal['keras', 'tqdm', 'progressbar2']='keras', progbar_width:int=None,
                stateful_metrics:Union[str, Set[str], Tuple[str], List[str]]=None, smooth_interval:int=100, **kwargs):
        '''complile: 定义loss, optimizer, metrics等参数
        
        :param loss: loss
        :param optimizer: 优化器
        :param scheduler: lr_scheduler
        :param clip_grad_norm: float, 是否使用梯度裁剪, 默认为False
        :param mixed_precision: bool, 是否使用混合精度, 默认为False
        :param metrics: str/List[str]/dict, 训练过程中需要打印的指标, loss相关指标默认会打印, 目前支持accuracy, 也支持自定义metric, 形式为{key: func}
        :param grad_accumulation_steps: int, 梯度累积步数, 默认为1
        :param bar: str, 使用进度条的种类, 从kwargs中解析, 默认为keras, 可选keras, tqdm, progressbar2

        > 进度条的配置
            progbar_type: str, 默认为keras, 可选keras, tqdm, progressbar2
            width: int, keras进度条下表示进度条的长度
        > 指标平滑的配置, 默认为None表示采取默认平滑设置; 传入False表示不使用平滑
            stateful_metrics: List[str], 表示不使用指标平滑仅进行状态记录的metric, 指标抖动会更加明显, 默认为None表示使用指标平滑
            smooth_interval: int, 表示指标平滑时候的累计步数, 默认为100

        :return: None
        '''
        self.criterion = loss or self.criterion  # loss
        self.optimizer = optimizer or self.optimizer  # 优化器
        self.scheduler = scheduler or self.scheduler  # lr_scheduler
        self.clip_grad_norm = clip_grad_norm  # 梯度裁剪
        self.grad_accumulation_steps = grad_accumulation_steps  # 梯度累积

        # 混合精度
        assert mixed_precision in {True, False, 'fp16', 'bf16'}
        self.mixed_precision_mode = 'fp16' if mixed_precision is True else mixed_precision
        if self.mixed_precision_mode:
            self.autocast = torch.cuda.amp.autocast
            self.scaler = torch.cuda.amp.GradScaler()

        # 训练过程观测的指标
        self.metrics = OrderedDict({'loss': None})
        if metrics is None:
            metrics = []
        elif isinstance(metrics, (str, dict)) or isfunction(metrics):
            metrics = [metrics]
        for metric in metrics:
            # 字符类型, 目前仅支持accuracy
            if isinstance(metric, str) and metric != 'loss':
                self.metrics[metric] = None
            # 字典形式 {metric: func}
            elif isinstance(metric, dict):
                self.metrics.update(metric)
            # 函数形式, key和value都赋值metric
            elif isfunction(metric):
                self.metrics.update({metric: metric})
            else:
                raise TypeError('Args metrics only support `String, Dict, Callable, List[String, Dict, Callable]` format')

        # 进度条参数
        assert progbar_type in {'keras', 'tqdm', 'progressbar2'}
        self.progbar_config = {'bar': progbar_type, 'width': progbar_width}
        self.progbar_config = {k:v for k,v in self.progbar_config.items() if v is not None}

        # smooth_metrics参数: 默认平滑
        self.smooth_metrics_config = {'stateful_metrics': stateful_metrics, 'interval': smooth_interval, 'verbose': kwargs.get('verbose')}
        self.smooth_metrics_config = {k:v for k,v in self.smooth_metrics_config.items() if v is not None}

        # 其他参数设置
        for key, value in kwargs.items():
            if hasattr(self, key):
                setattr(self, key, value)

    def print_trainable_parameters(self):
        '''打印可训练的参数量'''
        print_trainable_parameters(self.unwrap_model())

    @property
    def device(self) -> torch.device:
        '''获取model所在的device'''
        if hasattr(self, '_device'):
            return self._device
        return get_parameter_device(self.unwrap_model())

    @device.setter
    def device(self, value):
        '''允许修改self.device'''
        self._device = value

    def _move_to_model_device(self, inputs:Union[torch.Tensor, tuple, list, dict]):
        '''遍历并转移到model.device上（递归）'''
        if self.move_to_model_device:
            if isinstance(inputs, torch.Tensor) and hasattr(self, 'device') and (inputs.device != self.device):
                inputs = inputs.to(self.device)
            elif isinstance(inputs, (tuple, list)):
                inputs = list(inputs) if isinstance(inputs, tuple) else inputs
                for i, ts in enumerate(inputs):
                    inputs[i] = self._move_to_model_device(ts)
            elif isinstance(inputs, dict):
                for k, v in inputs.items():
                    inputs[k] = self._move_to_model_device(v)
        return inputs

    def _log_first_step(self, resume_step, train_X, train_y):
        '''打印第一个step的数据'''
        if self.log_first_step and self.verbose and (self.global_step == resume_step):
            print(colorful('[Train_data]: ', color='green'), + train_X)
            print(colorful('[Label]: ', color='green'), + train_y)

    def _forward(self, *inputs, **input_kwargs):
        '''调用模型的forward, 方便下游继承的时候可以自定义使用哪个模型的forward
        '''
        return self._argparse_forward(self.unwrap_model(), *inputs, **input_kwargs)

    def _argparse_forward(self, model:nn.Module, *inputs, **input_kwargs):
        '''调用模型的forward
        如果传入了网络结构module, 则调用module的forward; 如果是继承方式, 则调用自身的forward
        这里声明为staticmethod的话, 使用add_trainer会有问题
        '''
        if (len(inputs)==1) and isinstance(inputs[0], (tuple,list)):  # 防止([])嵌套
            inputs = inputs[0]
        
        if isinstance(inputs, torch.Tensor):  # tensor不展开
            return model.forward(inputs, **input_kwargs)
        elif isinstance(inputs, (tuple, list)):
            return model.forward(*inputs, **input_kwargs)
        else:
            return model.forward(inputs, **input_kwargs)

    def train_step(self, train_X, train_y):
        ''' Perform a training step on a batch of inputs. '''
        # 计算loss
        if self.mixed_precision_mode:
            with self.autocast(dtype=torch.float16 if self.mixed_precision_mode=='fp16' else torch.bfloat16):
                output = self._forward(train_X)
                loss_detail = self.criterion(output, train_y)
        else:
            output = self._forward(train_X)
            loss_detail = self.criterion(output, train_y)

        # 整理loss
        if isinstance(loss_detail, torch.Tensor):
            loss = loss_detail
            loss_detail = {}
        elif isinstance(loss_detail, dict):
            loss = loss_detail['loss']  # 还存在其他loss, 仅用于打印
            del loss_detail['loss']
        elif isinstance(loss_detail, (tuple, list)):
            loss = loss_detail[0]
            loss_detail = {f'loss{i}':v for i, v in enumerate(loss_detail[1:], start=1)}
        else:
            raise ValueError('Return loss only support `Tensor/dict/tuple/list` format')

        # 梯度累积
        loss = loss / self.grad_accumulation_steps if self.grad_accumulation_steps > 1 else loss

        # loss backward
        loss = self.loss_backward(loss)
        loss_detail = {k: (v.item() if isinstance(v, torch.Tensor) else v) / self.grad_accumulation_steps for k, v in loss_detail.items()}
        return output, loss, loss_detail

    def loss_backward(self, loss):
        '''loss.backward'''
        self.scale_before_step = 0
        if self.mixed_precision_mode:  # 混合精度
            self.scale_before_step = self.scaler.get_scale()
            self.scaler.scale(loss).backward(retain_graph=self.retain_graph)
        else:
            loss.backward(retain_graph=self.retain_graph)
        return loss
    
    def step(self):
        '''参数更新'''
        skip_scheduler = False
        # 混合精度
        if self.mixed_precision_mode:
            self.scaler.unscale_(self.optimizer)
            if self.clip_grad_norm is not None:  # 梯度裁剪
                torch.nn.utils.clip_grad_norm_(self.unwrap_model().parameters(), self.clip_grad_norm)
            self.scaler.step(self.optimizer)
            self.scaler.update()
            skip_scheduler = self.scaler.get_scale() != self.scale_before_step
        else:
            if self.clip_grad_norm is not None:  # 梯度裁剪
                torch.nn.utils.clip_grad_norm_(self.unwrap_model().parameters(), self.clip_grad_norm)
            self.optimizer.step()

        self.optimizer.zero_grad()  # 清梯度
        if (self.scheduler is not None) and not skip_scheduler:
            if isinstance(self.scheduler, (tuple, list)):
                for scheduler in self.scheduler:
                    scheduler.step()
            else:
                self.scheduler.step()

    def _prepare_inputs(self, train_dataloader:DataLoader, steps_per_epoch:Union[int,None], epochs:int, verbose:int):
        '''对fit的输入进行类型检查并置为成员变量'''
        if not hasattr(train_dataloader, '__len__'):
            assert steps_per_epoch is not None, 'Either train_dataloader has attr `__len__` or steps_per_epoch is not None'
        if steps_per_epoch is None:
            self.steps_per_epoch = math.ceil(len(train_dataloader) // self.grad_accumulation_steps)
        else:
            self.steps_per_epoch = steps_per_epoch
        self.batch_size = train_dataloader.batch_size
        self.epochs = epochs
        self.total_steps = self.steps_per_epoch * epochs
        self.train_dataloader = train_dataloader  # 设置为成员变量, 可由外部的callbacks进行修改
        self.train_dataloader_iter = iter(self.train_dataloader)  # 循环epoch时不重生成
        self.verbose = self.verbose if hasattr(self, 'verbose') else verbose

    def _prepare_callbacks(self, callbacks:Union[Callback, List[Callback]]=None):
        '''callbacks设置'''
        if callbacks is None:
            callbacks = []
        elif isinstance(callbacks, Callback):
            callbacks = [callbacks]
        for callback in callbacks:
            assert isinstance(callback, Callback), f"Args `callbacks` only support Callback(), but got {type(callback)}"

        history = History()
        callbacks_ = []

        # 指标平滑
        if any([isinstance(i, SmoothMetricsCallback) for i in callbacks]):
            # 用户自定的callbacks中包含了SmoothMetricsCallback
            log_warn(f'SmoothMetricsCallback already in use and args `smooth_metrics_config` will be ignored')
            smooth_callback = [callback for callback in callbacks if isinstance(callback, SmoothMetricsCallback)][0]
            callbacks_.append(smooth_callback)
            callbacks = [callback for callback in callbacks if not isinstance(callback, SmoothMetricsCallback)]
        elif self.smooth_metrics_config.get('interval') is not None:
            self.smooth_metrics_config['seen_so_far'] = self.resume_step + self.resume_epoch * self.steps_per_epoch  # 支持断点续训
            smooth_callback = SmoothMetricsCallback(**self.smooth_metrics_config)
            callbacks_.append(smooth_callback)
        else:
            # 不平滑
            smooth_callback = None

        # 检查指标平滑的设置和后续callback的设置的interval是不是一致
        for callback in callbacks:
            if hasattr(callback, 'interval') and (smooth_callback is not None) and (callback != smooth_callback) and \
                (callback.interval is not None) and (callback.interval % smooth_callback.interval != 0):
                log_warn(f'{type(callback).__name__}.interval={callback.interval} while SmoothMetricsCallback.interval={smooth_callback.interval}')
        
        # 进度条
        progbarlogger = None
        if self.verbose:
            if self.progbar_config['bar'] == 'keras':
                progbarlogger = KerasProgbar(**self.progbar_config)
            elif self.progbar_config['bar'] == 'tqdm':
                progbarlogger = TqdmProgbar(**self.progbar_config)
            elif self.progbar_config['bar'] == 'progressbar2':
                progbarlogger = ProgressBar2Progbar(**self.progbar_config)
            else:
                progbarlogger = KerasProgbar(**self.progbar_config)
            callbacks_.append(progbarlogger)

        callbacks_  += callbacks + [history]
        self.callbacks = CallbackList(callbacks_, run_callbacks=self.run_callbacks)
        callback_trainer = self
        callback_model = self.unwrap_model()
        params = {
            'epochs': self.epochs,
            'steps': self.steps_per_epoch,
            'verbose': self.verbose,
            'metrics': [i for i in self.metrics.keys() if isinstance(i, str)],
        }
        self.callbacks.set_all(trainer=callback_trainer, model=callback_model, optimizer=self.optimizer, scheduler=self.scheduler, params=params)
        callback_trainer.stop_training = False  # 在EarlyStopping中会重新设置
        return history, callback_trainer, progbarlogger

    def prepare_nextbatch(self):
        '''准备下一个batch数据'''
        # 循环dataloader, 不要试用itertools的cycle, 遇到过变量不释放的问题
        def _prepare_nextbatch():
            try:
                batch = next(self.train_dataloader_iter)
            except StopIteration:
                self.callbacks.on_dataloader_end()  # 适用于数据量较大时, 动态读取文件并重新生成self.train_dataloader的情况, 如预训练
                # DDP训练时候为了避免每个epoch样本一致, 修改随机种子
                if isinstance(self.train_dataloader.sampler, torch.utils.data.distributed.DistributedSampler) and \
                    hasattr(self.train_dataloader.sampler, 'set_epoch'):
                    self.train_dataloader.sampler.set_epoch(self.epoch)
                self.train_dataloader_iter = iter(self.train_dataloader)  # shuffle=True时候, 其实顺序也重新生成了
                batch = next(self.train_dataloader_iter)
            self.batch_step += 1
            return batch
        
        # 若使用梯度累计grad_accumulation_steps，则batch_step和global_step会不一致
        if self.batch_step > self.resume_batch:
            # 从头开始
            batch = _prepare_nextbatch()
        else:
            # 断点续训
            while self.batch_step <= self.resume_batch:
                batch = _prepare_nextbatch()
        batch = self._move_to_model_device(batch)
        return batch

    def fit(self, train_dataloader:DataLoader, steps_per_epoch:int=None, epochs:int=1, 
            callbacks:Union[Callback, List[Callback]]=None, verbose:int=1, **kwargs):
        ''' 模型训练
        :param train_dataloader: Dataloader, 训练数据集
        :param steps_per_epoch: int, 每个epoch训练的steps, 默认为None表示自行计算 
        :param epochs: int, 训练的轮次, 默认为1
        :param callbacks: Callback/List[Callback], 回调函数, 可调用预制的Callback或者自定义, 默认为None 
        :param verbose: int, 是否打印, 默认为1表示打印
        
        > 其他参数
        :param mail_receivers_when_error: str, 训练异常时候邮件通知
        :param save_ckpt_dir_when_error: str, 训练异常时候保存权重的路径
        :param save_batch_path_when_error: bool, 训练异常时候保存当前batch, 方便debug

        :return: History
        '''
        try:
            return self._fit(train_dataloader, steps_per_epoch, epochs, callbacks, verbose, **kwargs)
        except Exception as e:
            # 训练异常则发邮件
            error_msg = traceback.format_exc()
            mail_receivers_ = kwargs.get('mail_receivers_when_error')
            if mail_receivers_ is not None:
                mail_subject_ = kwargs.get('mail_subject_when_error') or "[ERROR] fit"
                mail_host_ = kwargs.get('mail_host_when_error')
                mail_user_ = kwargs.get('mail_user_when_error')
                mail_pwd_ = kwargs.get('mail_pwd_when_error')
                mail_sender_ = kwargs.get('mail_sender_when_error')
                send_email(mail_receivers_, mail_subject_, error_msg, mail_host=mail_host_, 
                           mail_user=mail_user_, mail_pwd=mail_pwd_, mail_sender=mail_sender_)

            # 训练异常则保存权重
            if kwargs.get('save_ckpt_dir_when_error') is not None:
                self.save_to_checkpoint(kwargs['save_ckpt_dir_when_error'], verbose=verbose, **kwargs)

            # 训练异常则打印当前batch
            save_batch_path_when_error = kwargs.get('save_batch_path_when_error')
            if save_batch_path_when_error is not None:
                os.makedirs(os.path.dirname(save_batch_path_when_error), exist_ok=True)
                torch.save({'train_X': self.train_X.cpu(), 'train_y': self.train_y.cpu()}, save_batch_path_when_error)
            
            raise e

    def _fit(self, train_dataloader:DataLoader, steps_per_epoch:int=None, epochs:int=1, 
             callbacks:Union[Callback, List[Callback]]=None, verbose:int=1, **kwargs):
        '''模型训练'''
        # 输入处理
        self._prepare_inputs(train_dataloader, steps_per_epoch, epochs, verbose)

        # 准备callbacks
        history, callback_trainer, progbarlogger  = self._prepare_callbacks(callbacks)

        #       epoch: 当前epoch
        # global_step: 当前全局训练步数
        #  local_step: 当前epoch内的训练步数, 不同epoch中相同local_step对应的batch数据不一定相同, 在steps_per_epoch=None时相同
        #  batch_step: 在dataloader中的index, 不同epoch中相同的bti对应的batch数据一般相同, 除非重新生成dataloader
        self.callbacks.on_train_begin()
        logs = self._log_init()  # 防止数据集为空时候
        for epoch in range(self.resume_epoch, epochs):
            self.epoch = epoch
            # resume_step：判断local_step的起点, 以及进度条的起始位置
            resume_step = self.resume_step if epoch==self.resume_epoch else 0
            self.callbacks.on_epoch_begin(self.global_step, self.epoch)
            if self.verbose:
                progbarlogger.seen = resume_step  # 这里设置进度条的seen, 在callbacks中也会修改
            
            for local_step in range(resume_step, self.steps_per_epoch):
                self.local_step = local_step
                self.global_step = self.epoch * self.steps_per_epoch + self.local_step
                logs = self._log_init()
                self.callbacks.on_batch_begin(self.global_step, self.local_step, logs)

                # forward和backward
                if not self.unwrap_model().training:
                    self.unwrap_model().train()  # 设置为train模式
                tr_loss, tr_loss_detail = 0, {}
                for _ in range(self.grad_accumulation_steps):
                    self.train_X, self.train_y = self.prepare_nextbatch()  # 获取下一个batch的训练数据
                    self._log_first_step(resume_step, self.train_X, self.train_y)  # log第一个step
                    output, loss, loss_detail = self.train_step(self.train_X, self.train_y)
                    self.callbacks.on_train_step_end()
                    tr_loss += loss.item()
                    for k, v in loss_detail.items():
                        tr_loss_detail[k] = tr_loss_detail.get(k, 0) + v
                # TODO: 理论上梯度累积时需对output和train_y进行合并, 主要是为了metric_mapping计算的准确
                
                # 参数更新
                self.step()

                # 添加loss至log打印
                logs.update(dict({'loss': tr_loss}, **tr_loss_detail))
                if self.verbose and self.loss2metrics and (self.global_step == resume_step):
                    # 把loss_detail添加到进度条metrics中
                    progbarlogger.add_metrics(list(tr_loss_detail.keys()), add_position=1)
                    
                # 添加metrics至log打印
                for metric, func in self.metrics.items():
                    perf = metric_mapping(metric, func, output, self.train_y)  # 内置的一些accuracy指标
                    if perf is not None:
                        if isfunction(metric):  # 直接传入回调函数(无key)
                            if self.verbose and (self.global_step == resume_step):
                                progbarlogger.add_metrics(list(perf.keys()))
                            logs.update(perf)
                        elif isinstance(metric, str):  # 直接传入回调函数(有key)
                            logs[metric] = perf

                self.callbacks.on_batch_end(self.global_step, self.local_step, logs)

            self.callbacks.on_epoch_end(self.global_step, self.epoch, logs)
            # TerminateOnNaN、EarlyStopping等停止训练策略
            if callback_trainer.stop_training:
                break
        self.callbacks.on_train_end(logs)
        return history

    def _log_init(self):
        '''获取batch_size, 主要是用于callback中的BaseLogger和Callback
        '''
        logs = {}

        # 添加lr
        try:
            logs['lr'] = self.optimizer.param_groups[0]["lr"]
        except:
            pass
        return logs

    @torch.no_grad()
    def predict(self, *inputs, **input_kwargs):
        '''模型预测, 调用forward()'''
        self.unwrap_model().eval()
        inputs = self._move_to_model_device(inputs)
        input_kwargs = self._move_to_model_device(input_kwargs)
        return self._forward(*inputs, **input_kwargs)
        
    def load_steps_params(self, save_path:str):
        '''导入训练过程参数
        
        :param save_path: str, 训练过程参数保存路径
        '''
        step_params = torch.load(save_path)
        self.resume_step = step_params['resume_step'] 
        self.resume_epoch = step_params['resume_epoch']
        self.resume_batch = step_params.get('resume_batch', 0)  # 兼容老版本
        return step_params

    def save_steps_params(self, save_path:str):
        '''保存训练过程参数

        :param save_path: str, 训练过程参数保存路径
        '''
        step_params = {
            'resume_step': (self.local_step+1) % self.steps_per_epoch,  # 当前epoch下的step数量
            'resume_epoch': self.epoch + (self.local_step+1) // self.steps_per_epoch,  # 经过的epoch数量
            'resume_batch': self.batch_step  # 经过的batch数量
            }
        save_dir = os.path.dirname(save_path)
        os.makedirs(save_dir, exist_ok=True)
        torch.save(step_params, save_path)

    def load_weights(self, load_path:Union[str,tuple,list], strict:bool=True, mapping:Union[dict,Callable]=None):
        '''加载模型权重, 支持加载权重文件list

        :param save_path: str/tuple/list, 权重加载路径
        :param strict: bool, torch.load()是否严格加载
        :param mapping: dict/func, 指定key的映射
            1. mapping=None, 表示按照模型自身结构加载, 一般加载finetune后使用save_weights()保存出来的权重
            2. mapping自定义, 根据用户自定义mapping来加载权重
        '''
        if isinstance(load_path, (tuple, list)):
            strict = False  # 加载多个权重文件时候, strict设置为False
        elif isinstance(load_path, str):
            load_path = [load_path]
        else:
            raise ValueError('Args `load_path` only support str/tuple/list format')
        
        mapping = mapping or dict()
        for load_path_i in load_path:
            state_dict = load_checkpoint(load_path_i)
            for k in list(state_dict.keys()):
                if isinstance(mapping, dict) and k in mapping:
                    state_dict[mapping[k]] = state_dict.pop(k)
                elif isinstance(mapping, Callable):
                    state_dict[mapping(k)] = state_dict.pop(k)
            self.unwrap_model().load_state_dict(state_dict, strict=strict)

    def save_weights(self, save_path:str, mapping:Union[dict,Callable]=None, trainable_only:bool=False):
        '''保存模型权重

        :param save_path: str, 权重保存路径
        :param mapping: dict/func, 指定key的映射
            1. mapping=None, 表示按照模型自身结构的key来保存, 后续可直接load_weights()加载
            2. mapping自定义, 根据用户自定义mapping来保存权重
        :param trainable_only: bool, 指定仅保存可训练参数
        '''
        state_dict = self.unwrap_model().state_dict()
        trainable_parameters = set(p for p,v in self.unwrap_model().named_parameters() if v.requires_grad)
        
        mapping = mapping or dict()
        for k in list(state_dict.keys()):
            # 只保存可训练的模型部分
            if trainable_only and (k not in trainable_parameters):
                continue
            if isinstance(mapping, dict) and k in mapping:
                state_dict[mapping[k]] = state_dict.pop(k)
            elif isinstance(mapping, Callable):
                state_dict[mapping(k)] = state_dict.pop(k)        
        save_checkpoint(state_dict, save_path)
        if trainable_only:
            params_all = sum(p.numel() for p in self.unwrap_model().parameters())
            params_trainable = sum(p.numel() for p in self.unwrap_model().parameters() if p.requires_grad)
            ratio = params_trainable/params_all*100
            log_info(f"Only trainable parameters saved and occupy {params_trainable}/{params_all}={ratio:.2f}%")

    def save_pretrained(self, save_path:str, weight_map:dict=None, mapping:Union[dict,Callable]=None):
        '''按照预训练模型的key来保存模型, 可供transformers包加载

        :param save_path: str, 保存的文件/文件夹路径
        '''
        state_dict = dict()
        for name, child in self.unwrap_model().named_children():
            if (name != '') and hasattr(child, 'save_pretrained'):
                tmp = child.save_pretrained(save_path, weight_map, mapping, write_to_disk=False)
                state_dict.update(tmp if tmp else {})
            else:
                state_dict.update({f'{name}.{k}': v for k,v in child.state_dict().items()})
        if len(state_dict) > 0:
            save_dir = None if re.search(r'\.[a-zA-z0-9]+$', save_path) else save_path
            save_checkpoint(state_dict, os.path.join(save_dir, 'pytorch_model.bin') if save_dir else save_path)
    
    def resume_from_checkpoint(self, save_dir:str=None, mapping:Union[dict,Callable]=None, strict:bool=True, device=None, verbose:int=1, **kwargs):
        '''同时加载模型、优化器、训练过程参数, 以保证断点续训和不断效果一致或相当

        :param save_dir: str, 保存目录
        :param mapping: dict, 模型文件的mapping
        :param strict: bool, 是否严格加载
        :param device: 加载的device
        :param verbose: int, 是否打印resume成功的信息, 默认打印

        > 可选参数
        :param model_path: str, 模型文件路径
        :param optimizer_path: str, 优化器文件路径
        :param scheduler_path: str, scheduler文件路径
        :param steps_params_path: str, 训练过程参数保存路径
        '''
        model_path = kwargs.get('model_path')
        optimizer_path = kwargs.get('optimizer_path')
        scheduler_path = kwargs.get('scheduler_path')
        steps_params_path = kwargs.get('steps_params_path')

        resume_info = []
        # 加载模型权重
        if model_path or save_dir:
            model_path = model_path or os.path.join(save_dir, 'model.pt')
            self.load_weights(model_path, strict=strict, mapping=mapping)
            resume_info.append(['Model weights', model_path])

        # 加载优化器
        if optimizer_path or save_dir:
            optimizer_path = optimizer_path or os.path.join(save_dir, 'optimizer.pt')
            state_dict = torch.load(optimizer_path, map_location = device or self.device)
            self.optimizer.load_state_dict(state_dict)
            resume_info.append(['Optimizer', optimizer_path])

        # 加载scheduler
        if (scheduler_path or save_dir) and (self.scheduler is not None):
            scheduler_path = scheduler_path or os.path.join(save_dir, 'scheduler.pt')
            state_dict = torch.load(scheduler_path, map_location = device or self.device)
            self.scheduler.load_state_dict(state_dict)
            resume_info.append(['Scheduler', scheduler_path])

        # 加载训练进度参数
        if steps_params_path or save_dir:
            steps_params_path = steps_params_path or os.path.join(save_dir, 'steps_params.pt')
            self.load_steps_params(steps_params_path)
            resume_info.append(['Steps_params', steps_params_path])

        if verbose == 1 and len(resume_info) > 0:
            log_info('Successfuly resume training checkpoint')
            print_table(resume_info, headers=['File', 'Path'])

    def save_to_checkpoint(self, save_dir:str=None, mapping:Union[dict,Callable]=None, trainable_only:bool=False, verbose:int=0, **kwargs):
        '''同时保存模型、优化器、训练过程参数、scheduler

        :param save_dir: str, 保存目录
        :param mapping: dict/func, 模型文件的mapping
        :param trainable_only

        > 可选参数
        :param model_path: str, 模型文件路径
        :param optimizer_path: str, 优化器文件路径
        :param scheduler_path: str, scheduler文件路径
        :param steps_params_path: str, 训练过程参数保存路径
        '''
        model_path = kwargs.get('model_path')
        optimizer_path = kwargs.get('optimizer_path')
        scheduler_path = kwargs.get('scheduler_path')
        steps_params_path = kwargs.get('steps_params_path')

        load_info = []
        if model_path or save_dir:
            model_path = model_path or os.path.join(save_dir, 'model.pt')
            self.save_weights(model_path, mapping=mapping, trainable_only=trainable_only)
            load_info.append(['Model weights', model_path])

        if optimizer_path or save_dir:
            optimizer_path = optimizer_path or os.path.join(save_dir, 'optimizer.pt')
            os.makedirs(os.path.dirname(optimizer_path), exist_ok=True)
            torch.save(self.optimizer.state_dict(), optimizer_path)
            load_info.append(['Optimizer', optimizer_path])

        if (scheduler_path or save_dir) and (self.scheduler is not None):
            scheduler_path = scheduler_path or os.path.join(save_dir, 'scheduler.pt')
            os.makedirs(os.path.dirname(scheduler_path), exist_ok=True)
            torch.save(self.scheduler.state_dict(), scheduler_path)
            load_info.append(['Scheduler', scheduler_path])

        if steps_params_path or save_dir:
            steps_params_path = steps_params_path or os.path.join(save_dir, 'steps_params.pt')
            self.save_steps_params(steps_params_path)
            load_info.append(['Steps_params', steps_params_path])

        if verbose == 1 and len(load_info) > 0:
            log_info('Successfuly save training checkpoint')
            print_table(load_info, headers=['File', 'Path'])

    def unwrap_model(self) -> nn.Module:
        '''返回nn.Module模块
        '''
        if isinstance(self, nn.Module):
            return self
        elif hasattr(self, 'module') and isinstance(self.module, nn.Module):
            return self.module
        else:
            return self

    # 以下为增加了nn.Module自带的方法
    def register_buffer(self, *args, **kwargs):
        self.unwrap_model().register_buffer(*args, **kwargs)

    def register_parameter(self, *args, **kwargs):
        self.unwrap_model().register_parameter(*args, **kwargs)

    def add_module(self, *args, **kwargs):
        self.unwrap_model().add_module(*args, **kwargs)

    def register_module(self, *args, **kwargs):
        self.unwrap_model().register_module(*args, **kwargs)

    def get_submodule(self, *args, **kwargs):
        return self.unwrap_model().get_submodule(*args, **kwargs)

    def get_parameter(self, *args, **kwargs):
        return self.unwrap_model().get_parameter(*args, **kwargs)

    def get_buffer(self, *args, **kwargs):
        return self.unwrap_model().get_buffer(*args, **kwargs)

    def apply(self, *args, **kwargs):
        self.unwrap_model().apply(*args, **kwargs)
        return self

    def cuda(self, *args, **kwargs):
        self.unwrap_model().cuda(*args, **kwargs)
        return self

    def ipu(self, *args, **kwargs):
        self.unwrap_model().ipu(*args, **kwargs)
        return self

    def xpu(self, *args, **kwargs):
        self.unwrap_model().xpu(*args, **kwargs)
        return self

    def cpu(self, *args, **kwargs):
        self.unwrap_model().cpu(*args, **kwargs)
        return self

    def type(self, *args, **kwargs):
        self.unwrap_model().type(*args, **kwargs)
        return self

    def float(self, *args, **kwargs):
        self.unwrap_model().float(*args, **kwargs)
        return self

    def double(self, *args, **kwargs):
        self.unwrap_model().double(*args, **kwargs)
        return self

    def half(self, *args, **kwargs):
        self.unwrap_model().half(*args, **kwargs)
        return self

    def bfloat16(self, *args, **kwargs):
        self.unwrap_model().bfloat16(*args, **kwargs)
        return self

    def to_empty(self, *args, **kwargs):
        self.unwrap_model().to_empty(*args, **kwargs)
        return self

    def to(self, *args, **kwargs):
        self.unwrap_model().to(*args, **kwargs)
        return self

    def register_full_backward_pre_hook(self, *args, **kwargs):
        return self.unwrap_model().register_full_backward_pre_hook(*args, **kwargs)

    def register_backward_hook(self, *args, **kwargs):
        return self.unwrap_model().register_backward_hook(*args, **kwargs)

    def register_full_backward_hook(self, *args, **kwargs):
        return self.unwrap_model().register_full_backward_hook(*args, **kwargs)

    def register_forward_pre_hook(self, *args, **kwargs):
        return self.unwrap_model().register_forward_pre_hook(*args, **kwargs)

    def register_forward_hook(self, *args, **kwargs):
        return self.unwrap_model().register_forward_hook(*args, **kwargs)

    def register_state_dict_pre_hook(self, *args, **kwargs):
        return self.unwrap_model().register_state_dict_pre_hook(*args, **kwargs)

    def state_dict(self, *args, **kwargs):
        return self.unwrap_model().state_dict(*args, **kwargs)

    def register_load_state_dict_post_hook(self, *args, **kwargs):
        return self.unwrap_model().register_load_state_dict_post_hook(*args, **kwargs)

    def load_state_dict(self, *args, **kwargs):
        return self.unwrap_model().load_state_dict(*args, **kwargs)

    def parameters(self, *args, **kwargs):
        return self.unwrap_model().parameters(*args, **kwargs)

    def named_parameters(self, *args, **kwargs):
        return self.unwrap_model().named_parameters(*args, **kwargs)

    def buffers(self, *args, **kwargs):
        return self.unwrap_model().buffers(*args, **kwargs)

    def named_buffers(self, *args, **kwargs):
        return self.unwrap_model().named_buffers(*args, **kwargs)

    def children(self, *args, **kwargs):
        return self.unwrap_model().children(*args, **kwargs)

    def named_children(self, *args, **kwargs):
        return self.unwrap_model().named_children(*args, **kwargs)

    def modules(self, *args, **kwargs):
        return self.unwrap_model().modules(*args, **kwargs)

    def named_modules(self, *args, **kwargs):
        return self.unwrap_model().named_modules(*args, **kwargs)

    def train(self, *args, **kwargs):
        self.unwrap_model().train(*args, **kwargs)
        return self

    def eval(self, *args, **kwargs):
        self.unwrap_model().eval(*args, **kwargs)
        return self

    def requires_grad_(self, *args, **kwargs):
        self.unwrap_model().requires_grad_(*args, **kwargs)
        return self

    def zero_grad(self, *args, **kwargs):
        self.unwrap_model().zero_grad(*args, **kwargs)

    def share_memory(self, *args, **kwargs):
        self.unwrap_model().share_memory(*args, **kwargs)
        return self


Trainer.compile_training_components = Trainer.compile



from torch import nn
from torch.utils.data import DataLoader
from torch4keras.snippets import log_warn
from typing import Union
from .base import Trainer


class AccelerateTrainer(Trainer):
    '''accelerate来训练'''
    def __init__(self, module: nn.Module, **configs):
        super().__init__(module)
        from accelerate import Accelerator
        accelerator = Accelerator(**configs)
        self.model = accelerator.prepare(module)
        self.accelerator = accelerator
        self.device = accelerator.device
        self.verbose = 1 if accelerator.is_local_main_process else 0
        log_warn('AcclerateTrainer may not be compatible with several callbacks, you may use custom callbacks instead.')
    
    def compile(self, *args, **kwargs):
        super().compile(*args, **kwargs)
        self.optimizer, self.scheduler, self.criterion = self.accelerator.prepare(self.optimizer, self.scheduler, self.criterion)

    def _prepare_inputs(self, train_dataloader:DataLoader, steps_per_epoch:Union[int,None], epochs:int, verbose:int):
        # 如果使用ddp的时候没有使用DistributedSampler，这里会自动修改一下
        train_dataloader = self.accelerator.prepare(train_dataloader)
        super()._prepare_inputs(train_dataloader, steps_per_epoch, epochs, verbose)

    def prepare(self, *args, **kwargs):
        '''调用acclerate的prepare, 如在外面评估时候需要对dev_dataloader使用'''
        return self.accelerator.prepare(*args, **kwargs)

    def unwrap_model(self):
        '''返回nn.Module模块'''
        unwrap_model = self.accelerator.unwrap_model(self.model)
        if isinstance(unwrap_model, nn.Module): return unwrap_model
        return unwrap_model.module if hasattr(unwrap_model, 'module') else unwrap_model

    def loss_backward(self, loss):
        self.accelerator.backward(loss)
        return loss